==> dextrades-0.1.6.tar.gz: dextrades-0.1.6/src/event_pipeline.rs <==
/// Multi-stage event processing pipeline for RPC racing data integrity
/// 
/// Architecture: Ingestion -> Deduplication -> Reordering -> Enrichment -> Stream
/// 
/// This module implements the optimal solution combining insights from both Grok and Gemini:
/// - DashSet for concurrent deduplication with (tx_hash, log_index) keys
/// - BinaryHeap with watermark strategy for ordered release
/// - Decoupled enrichment with retry mechanisms
/// - Pruning to prevent unbounded memory growth

use crate::schema::SwapEvent;
use crate::service::DextradesService;
use dashmap::DashSet;
use std::collections::BinaryHeap;
use std::sync::Arc;
use std::time::{Duration, Instant};
use tokio::sync::RwLock;
use log::{debug, info, warn};

/// Unique identifier for a swap event: (tx_hash, log_index)
#[derive(Clone, Debug, Hash, PartialEq, Eq)]
pub struct EventId {
    pub tx_hash: String,
    pub log_index: u64,
}

/// Event with ordering information for the reordering buffer
#[derive(Debug, Clone)]
pub struct OrderedEvent {
    pub event: SwapEvent,
    pub block_number: u64,
    pub tx_index: Option<u64>, // For sub-block ordering
    pub log_index: u64,
}

impl PartialEq for OrderedEvent {
    fn eq(&self, other: &Self) -> bool {
        self.block_number == other.block_number 
            && self.tx_index == other.tx_index 
            && self.log_index == other.log_index
    }
}

impl Eq for OrderedEvent {}

impl PartialOrd for OrderedEvent {
    fn partial_cmp(&self, other: &Self) -> Option<std::cmp::Ordering> {
        Some(self.cmp(other))
    }
}

impl Ord for OrderedEvent {
    fn cmp(&self, other: &Self) -> std::cmp::Ordering {
        // Reverse order for min-heap behavior (BinaryHeap is max-heap by default)
        other.block_number.cmp(&self.block_number)
            .then_with(|| other.tx_index.cmp(&self.tx_index))
            .then_with(|| other.log_index.cmp(&self.log_index))
    }
}

/// Configuration for the event processing pipeline
#[derive(Debug, Clone)]
pub struct PipelineConfig {
    /// Buffer window size for reordering (in blocks)
    pub reorder_window_blocks: u64,
    /// Maximum time to wait before releasing events (latency vs completeness trade-off)
    pub max_reorder_delay: Duration,
    /// Pruning interval for the deduplication set
    pub pruning_interval: Duration,
}

impl Default for PipelineConfig {
    fn default() -> Self {
        Self {
            reorder_window_blocks: 10,  // 10 block buffer for ordering
            max_reorder_delay: Duration::from_secs(30),  // Max 30s delay
            pruning_interval: Duration::from_secs(300),  // Prune every 5 minutes
        }
    }
}

/// Multi-stage event processing pipeline
pub struct EventPipeline {
    config: PipelineConfig,
    
    // Stage 1: Deduplication
    seen_events: Arc<DashSet<EventId>>,
    
    // Stage 2: Reordering buffer
    reorder_buffer: Arc<RwLock<BinaryHeap<OrderedEvent>>>,
    highest_seen_block: Arc<RwLock<u64>>,
    
    // Pruning
    last_pruned: Arc<RwLock<Instant>>,
}

impl EventPipeline {
    /// Create a new event processing pipeline
    pub async fn new(config: PipelineConfig, _service: Arc<DextradesService>) -> Result<Self, Box<dyn std::error::Error + Send + Sync>> {
        Ok(Self {
            config,
            seen_events: Arc::new(DashSet::new()),
            reorder_buffer: Arc::new(RwLock::new(BinaryHeap::new())),
            highest_seen_block: Arc::new(RwLock::new(0)),
            last_pruned: Arc::new(RwLock::new(Instant::now())),
        })
    }
    
    /// Process a batch of events through the pipeline
    pub async fn process_events(&self, events: Vec<SwapEvent>) -> Vec<SwapEvent> {
        if events.is_empty() {
            return Vec::new();
        }
        
        info!("🔄 Processing {} events through pipeline", events.len());
        
        // Stage 1: Deduplication
        let unique_events = self.deduplicate_events(events).await;
        
        // Stage 2: Add to reordering buffer  
        self.add_to_reorder_buffer(unique_events).await;
        
        // Stage 3: Release ordered events when ready
        let ordered_events = self.release_ordered_events().await;
        
        // Stage 4: Pruning (periodic)
        self.maybe_prune().await;
        
        ordered_events
    }
    
    /// Stage 1: Deduplicate events using concurrent DashSet
    async fn deduplicate_events(&self, events: Vec<SwapEvent>) -> Vec<SwapEvent> {
        let mut unique_events = Vec::new();
        let mut duplicates = 0;
        
        info!("🔍 Deduplication: Processing {} input events", events.len());
        
        for event in events {
            let event_id = EventId {
                tx_hash: event.tx_hash.clone(),
                log_index: event.log_index,
            };
            
            // Atomic insert - returns true if new, false if duplicate
            if self.seen_events.insert(event_id.clone()) {
                debug!("✅ New event: {} log_index {} block {}", 
                       event_id.tx_hash, event_id.log_index, event.block_number);
                unique_events.push(event);
            } else {
                duplicates += 1;
                warn!("🔄 Filtered duplicate: {} log_index {} block {}", 
                      event_id.tx_hash, event_id.log_index, event.block_number);
            }
        }
        
        info!("✅ Deduplication result: {} unique, {} duplicates filtered (total seen: {})", 
              unique_events.len(), duplicates, self.seen_events.len());
        
        unique_events
    }
    
    /// Stage 2: Add events to reordering buffer
    async fn add_to_reorder_buffer(&self, events: Vec<SwapEvent>) {
        if events.is_empty() {
            return;
        }
        
        let mut buffer = self.reorder_buffer.write().await;
        let mut highest_block = self.highest_seen_block.write().await;
        
        for event in events {
            let block_number = event.block_number;
            
            // Update highest seen block
            if block_number > *highest_block {
                *highest_block = block_number;
            }
            
            let ordered_event = OrderedEvent {
                block_number,
                tx_index: None, // TODO: Extract from transaction if needed for sub-block ordering
                log_index: event.log_index,
                event,
            };
            
            buffer.push(ordered_event);
        }
        
        debug!("📦 Added events to reorder buffer, highest block: {}", *highest_block);
    }
    
    /// Stage 3: Release ordered events using watermark strategy
    async fn release_ordered_events(&self) -> Vec<SwapEvent> {
        let highest_block = *self.highest_seen_block.read().await;
        
        let mut buffer = self.reorder_buffer.write().await;
        let mut ready_events = Vec::new();
        
        // STREAMING FIX: If reorder_window_blocks is 0, release ALL events immediately for real-time streaming
        if self.config.reorder_window_blocks == 0 {
            // Immediate streaming mode - release all events in buffer
            while let Some(ordered_event) = buffer.pop() {
                ready_events.push(ordered_event.event);
            }
            
            // Sort by block number to maintain ordering
            ready_events.sort_by_key(|event| (event.block_number, event.log_index));
            
            if !ready_events.is_empty() {
                info!("⚡ Released {} events IMMEDIATELY for streaming (reorder_window=0)", ready_events.len());
            }
        } else {
            // Normal watermark strategy for buffered processing
            let watermark = highest_block.saturating_sub(self.config.reorder_window_blocks);
            
            // Pop events that are before the watermark (safely ordered)
            while let Some(ordered_event) = buffer.peek() {
                if ordered_event.block_number <= watermark {
                    let event = buffer.pop().unwrap();
                    ready_events.push(event.event);
                } else {
                    break;
                }
            }
            
            if !ready_events.is_empty() {
                debug!("🚀 Released {} ordered events (watermark: {}, highest: {})", ready_events.len(), watermark, highest_block);
            }
        }
        
        ready_events
    }
    
    /// Flush all remaining events from the reorder buffer (call at end of stream)
    pub async fn flush_remaining_events(&self) -> Vec<SwapEvent> {
        let mut buffer = self.reorder_buffer.write().await;
        let mut remaining_events = Vec::new();
        
        // Release all remaining events in order
        while let Some(ordered_event) = buffer.pop() {
            remaining_events.push(ordered_event.event);
        }
        
        // Sort by block number to maintain ordering
        remaining_events.sort_by_key(|event| event.block_number);
        
        if !remaining_events.is_empty() {
            info!("🔄 Flushed {} remaining events from reorder buffer", remaining_events.len());
        }
        
        remaining_events
    }
    
    /// Stage 4: Periodic pruning of old event IDs to prevent unbounded growth
    async fn maybe_prune(&self) {
        let now = Instant::now();
        let mut last_pruned = self.last_pruned.write().await;
        
        if now.duration_since(*last_pruned) >= self.config.pruning_interval {
            *last_pruned = now;
            drop(last_pruned);
            
            // Determine pruning threshold based on lowest watermark
            let highest_block = *self.highest_seen_block.read().await;
            let prune_threshold = highest_block.saturating_sub(self.config.reorder_window_blocks * 2);
            
            // TODO: Implement selective pruning based on block numbers
            // For now, we'll rely on DashSet's internal memory management
            // A production implementation would store block numbers with event IDs for selective pruning
            
            debug!("🧹 Pruning cycle completed (threshold: block {})", prune_threshold);
        }
    }
    
    /// Get current pipeline statistics
    pub async fn get_stats(&self) -> PipelineStats {
        let buffer = self.reorder_buffer.read().await;
        let highest_block = *self.highest_seen_block.read().await;
        
        PipelineStats {
            seen_events_count: self.seen_events.len(),
            reorder_buffer_size: buffer.len(),
            highest_seen_block: highest_block,
        }
    }
}

/// Pipeline statistics for monitoring
#[derive(Debug, Clone)]
pub struct PipelineStats {
    pub seen_events_count: usize,
    pub reorder_buffer_size: usize,
    pub highest_seen_block: u64,
}

==> dextrades-0.1.12.tar.gz: dextrades-0.1.12/src/event_pipeline.rs <==
/// Multi-stage event processing pipeline for RPC racing data integrity
/// 
/// Architecture: Ingestion -> Deduplication -> Reordering -> Enrichment -> Stream
/// 
/// This module implements the optimal solution combining insights from both Grok and Gemini:
/// - DashSet for concurrent deduplication with (tx_hash, log_index) keys
/// - BinaryHeap with watermark strategy for ordered release
/// - Decoupled enrichment with retry mechanisms
/// - Pruning to prevent unbounded memory growth

use crate::schema::SwapEvent;
use crate::service::DextradesService;
use dashmap::DashSet;
use std::collections::BinaryHeap;
use std::sync::Arc;
use std::time::{Duration, Instant};
use tokio::sync::RwLock;
use log::{debug, info, warn};

/// Unique identifier for a swap event: (tx_hash, log_index)
#[derive(Clone, Debug, Hash, PartialEq, Eq)]
pub struct EventId {
    pub tx_hash: String,
    pub log_index: u64,
}

/// Event with ordering information for the reordering buffer
#[derive(Debug, Clone)]
pub struct OrderedEvent {
    pub event: SwapEvent,
    pub block_number: u64,
    pub tx_index: Option<u64>, // For sub-block ordering
    pub log_index: u64,
}

impl PartialEq for OrderedEvent {
    fn eq(&self, other: &Self) -> bool {
        self.block_number == other.block_number 
            && self.tx_index == other.tx_index 
            && self.log_index == other.log_index
    }
}

impl Eq for OrderedEvent {}

impl PartialOrd for OrderedEvent {
    fn partial_cmp(&self, other: &Self) -> Option<std::cmp::Ordering> {
        Some(self.cmp(other))
    }
}

impl Ord for OrderedEvent {
    fn cmp(&self, other: &Self) -> std::cmp::Ordering {
        // Reverse order for min-heap behavior (BinaryHeap is max-heap by default)
        other.block_number.cmp(&self.block_number)
            .then_with(|| other.tx_index.cmp(&self.tx_index))
            .then_with(|| other.log_index.cmp(&self.log_index))
    }
}

/// Configuration for the event processing pipeline
#[derive(Debug, Clone)]
pub struct PipelineConfig {
    /// Buffer window size for reordering (in blocks)
    pub reorder_window_blocks: u64,
    /// Maximum time to wait before releasing events (latency vs completeness trade-off)
    pub max_reorder_delay: Duration,
    /// Pruning interval for the deduplication set
    pub pruning_interval: Duration,
}

impl Default for PipelineConfig {
    fn default() -> Self {
        Self {
            reorder_window_blocks: 10,  // 10 block buffer for ordering
            max_reorder_delay: Duration::from_secs(30),  // Max 30s delay
            pruning_interval: Duration::from_secs(300),  // Prune every 5 minutes
        }
    }
}

/// Multi-stage event processing pipeline
pub struct EventPipeline {
    config: PipelineConfig,
    
    // Stage 1: Deduplication
    seen_events: Arc<DashSet<EventId>>,
    
    // Stage 2: Reordering buffer
    reorder_buffer: Arc<RwLock<BinaryHeap<OrderedEvent>>>,
    highest_seen_block: Arc<RwLock<u64>>,
    
    // Pruning
    last_pruned: Arc<RwLock<Instant>>,
    // Time-based fallback release to mitigate late arrivals indefinitely blocking output
    last_release: Arc<RwLock<Instant>>,
}

impl EventPipeline {
    /// Create a new event processing pipeline
    pub async fn new(config: PipelineConfig, _service: Arc<DextradesService>) -> Result<Self, Box<dyn std::error::Error + Send + Sync>> {
        Ok(Self {
            config,
            seen_events: Arc::new(DashSet::new()),
            reorder_buffer: Arc::new(RwLock::new(BinaryHeap::new())),
            highest_seen_block: Arc::new(RwLock::new(0)),
            last_pruned: Arc::new(RwLock::new(Instant::now())),
            last_release: Arc::new(RwLock::new(Instant::now())),
        })
    }
    
    /// Process a batch of events through the pipeline
    pub async fn process_events(&self, events: Vec<SwapEvent>) -> Vec<SwapEvent> {
        if events.is_empty() {
            return Vec::new();
        }
        
        info!("🔄 Processing {} events through pipeline", events.len());
        
        // Stage 1: Deduplication
        let unique_events = self.deduplicate_events(events).await;
        
        // Stage 2: Add to reordering buffer  
        self.add_to_reorder_buffer(unique_events).await;
        
        // Stage 3: Release ordered events when ready
        let ordered_events = self.release_ordered_events().await;
        
        // Stage 4: Pruning (periodic)
        self.maybe_prune().await;
        
        ordered_events
    }
    
    /// Stage 1: Deduplicate events using concurrent DashSet
    async fn deduplicate_events(&self, events: Vec<SwapEvent>) -> Vec<SwapEvent> {
        let mut unique_events = Vec::new();
        let mut duplicates = 0;
        
        info!("🔍 Deduplication: Processing {} input events", events.len());
        
        for event in events {
            let event_id = EventId {
                tx_hash: event.tx_hash.clone(),
                log_index: event.log_index,
            };
            
            // Atomic insert - returns true if new, false if duplicate
            if self.seen_events.insert(event_id.clone()) {
                debug!("✅ New event: {} log_index {} block {}", 
                       event_id.tx_hash, event_id.log_index, event.block_number);
                unique_events.push(event);
            } else {
                duplicates += 1;
                warn!("🔄 Filtered duplicate: {} log_index {} block {}", 
                      event_id.tx_hash, event_id.log_index, event.block_number);
            }
        }
        
        info!("✅ Deduplication result: {} unique, {} duplicates filtered (total seen: {})", 
              unique_events.len(), duplicates, self.seen_events.len());
        
        unique_events
    }
    
    /// Stage 2: Add events to reordering buffer
    async fn add_to_reorder_buffer(&self, events: Vec<SwapEvent>) {
        if events.is_empty() {
            return;
        }
        
        let mut buffer = self.reorder_buffer.write().await;
        let mut highest_block = self.highest_seen_block.write().await;
        
        for mut event in events {
            let block_number = event.block_number;
            // Extract tx_index for improved intra-block ordering
            let tx_index = event.get_enriched_u64("tx_index");
            // Also persist into the event for downstream consumers
            if event.tx_index.is_none() {
                event.tx_index = tx_index;
            }
            
            // Update highest seen block
            if block_number > *highest_block {
                *highest_block = block_number;
            }
            
            let ordered_event = OrderedEvent {
                block_number,
                tx_index,
                log_index: event.log_index,
                event,
            };
            
            buffer.push(ordered_event);
        }
        
        debug!("📦 Added events to reorder buffer, highest block: {}", *highest_block);
    }
    
    /// Stage 3: Release ordered events using watermark strategy
    async fn release_ordered_events(&self) -> Vec<SwapEvent> {
        let highest_block = *self.highest_seen_block.read().await;
        let now = Instant::now();

        let mut buffer = self.reorder_buffer.write().await;
        let mut ready_events = Vec::new();

        // STREAMING FIX: If reorder_window_blocks is 0, release ALL events immediately for real-time streaming
        if self.config.reorder_window_blocks == 0 {
            // Immediate streaming mode - release all events in buffer
            while let Some(ordered_event) = buffer.pop() {
                ready_events.push(ordered_event.event);
            }
            
            // Sort by block, tx_index, log_index to maintain ordering
            ready_events.sort_by(|a, b| {
                a.block_number
                    .cmp(&b.block_number)
                    .then_with(|| a.tx_index.cmp(&b.tx_index))
                    .then_with(|| a.log_index.cmp(&b.log_index))
            });
            
            if !ready_events.is_empty() {
                info!("⚡ Released {} events IMMEDIATELY for streaming (reorder_window=0)", ready_events.len());
                *self.last_release.write().await = now;
            }
        } else {
            // Normal watermark strategy for buffered processing
            let watermark = highest_block.saturating_sub(self.config.reorder_window_blocks);
            
            // Pop events that are before the watermark (safely ordered)
            while let Some(ordered_event) = buffer.peek() {
                if ordered_event.block_number <= watermark {
                    let event = buffer.pop().unwrap();
                    ready_events.push(event.event);
                } else {
                    break;
                }
            }
            
            if ready_events.is_empty() && !buffer.is_empty() {
                // Time-based fallback: if we have been waiting longer than max_reorder_delay,
                // release everything we have (sorted) to avoid long stalls in small ranges.
                let last = *self.last_release.read().await;
                if now.duration_since(last) >= self.config.max_reorder_delay {
                    while let Some(ordered_event) = buffer.pop() {
                        ready_events.push(ordered_event.event);
                    }
                    // Sort by block, tx_index, log_index
                    ready_events.sort_by(|a, b| {
                        a.block_number
                            .cmp(&b.block_number)
                            .then_with(|| a.tx_index.cmp(&b.tx_index))
                            .then_with(|| a.log_index.cmp(&b.log_index))
                    });
                    info!("⏳ Fallback release {} events after {:?} (watermark: {}, highest: {})",
                        ready_events.len(), self.config.max_reorder_delay, watermark, highest_block);
                }
            }

            if !ready_events.is_empty() {
                debug!("🚀 Released {} ordered events (watermark: {}, highest: {})", ready_events.len(), watermark, highest_block);
                *self.last_release.write().await = now;
            }
        }
        
        ready_events
    }
    
    /// Flush all remaining events from the reorder buffer (call at end of stream)
    pub async fn flush_remaining_events(&self) -> Vec<SwapEvent> {
        let mut buffer = self.reorder_buffer.write().await;
        let mut remaining_events = Vec::new();
        
        // Release all remaining events in order
        while let Some(ordered_event) = buffer.pop() {
            remaining_events.push(ordered_event.event);
        }
        
        // Sort by block, tx_index, log_index to maintain ordering
        remaining_events.sort_by(|a, b| {
            a.block_number
                .cmp(&b.block_number)
                .then_with(|| a.tx_index.cmp(&b.tx_index))
                .then_with(|| a.log_index.cmp(&b.log_index))
        });
        
        if !remaining_events.is_empty() {
            info!("🔄 Flushed {} remaining events from reorder buffer", remaining_events.len());
        }
        
        remaining_events
    }
    
    /// Stage 4: Periodic pruning of old event IDs to prevent unbounded growth
    async fn maybe_prune(&self) {
        let now = Instant::now();
        let mut last_pruned = self.last_pruned.write().await;
        
        if now.duration_since(*last_pruned) >= self.config.pruning_interval {
            *last_pruned = now;
            drop(last_pruned);
            
            // Determine pruning threshold based on lowest watermark
            let highest_block = *self.highest_seen_block.read().await;
            let prune_threshold = highest_block.saturating_sub(self.config.reorder_window_blocks * 2);
            
            // TODO: Implement selective pruning based on block numbers
            // For now, we'll rely on DashSet's internal memory management
            // A production implementation would store block numbers with event IDs for selective pruning
            
            debug!("🧹 Pruning cycle completed (threshold: block {})", prune_threshold);
        }
    }
    
    /// Get current pipeline statistics
    pub async fn get_stats(&self) -> PipelineStats {
        let buffer = self.reorder_buffer.read().await;
        let highest_block = *self.highest_seen_block.read().await;
        
        PipelineStats {
            seen_events_count: self.seen_events.len(),
            reorder_buffer_size: buffer.len(),
            highest_seen_block: highest_block,
        }
    }
}

/// Pipeline statistics for monitoring
#[derive(Debug, Clone)]
pub struct PipelineStats {
    pub seen_events_count: usize,
    pub reorder_buffer_size: usize,
    pub highest_seen_block: u64,
}
